#include "include.h"
#include "log.h"
#include "socket.h"
#include "socket_msg.h"

static struct uloop_fd server;
struct in6_addr ownaddr;
struct sockaddr_in6 client_addr;

int socket_info(struct socket_msg *info) {
	int ret = 0;
	char msg[REQUEST_MAXLEN];

	ret = socket_msg_marshal(info, msg);
	if (ret) {
		log_error("socket_request: could not marshel message: %d\n", ret);
		return ret;
	}

	ret = sendto(server.fd, msg, REQUEST_MAXLEN, 0, (struct sockaddr *)&client_addr, sizeof(client_addr));
	if (ret < 0) {
		log_error("socket_request: could not send message: %d\n", ret);
	}
	return ret;
}

int socket_request_timeout(struct socket_msg *request, struct socket_msg *answer, struct timeval timeout) {
	int ret = 0;
	int sock = 0;
	char msg[REQUEST_MAXLEN];
	request->type = (request->type | SOCKET_MSG_TYPE_REQUEST);

	ret = socket_msg_marshal(request, msg);
	if (ret) {
		log_error("socket_request: could not marshel message: %d\n", ret);
		return ret;
	}

	sock = socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
	if (sock < 0) {
		log_error("socket_request: could not create socket\n");
		return -1;
	}


	int loop = 0;
	if(setsockopt(sock, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, &loop, sizeof(loop)) < 0) {
		log_error("socket_request: unable to disable listen/loop own multicast package: %d\n", errno);
		return -1;
	}

	ret = sendto(sock, msg, REQUEST_MAXLEN, 0, (struct sockaddr *)&client_addr, sizeof(client_addr));
	if (ret < 0) {
		close(sock);
		log_error("socket_request: could not send message: %d\n", ret);
		return ret;
	}

	ret = setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (char*)&timeout,sizeof(timeout));
	if(ret < 0) {
		close(sock);
		log_error("socket_request: unable to set timeout: %d\n", errno);
		return ret;
	}

	ret = recv(sock, msg, REQUEST_MAXLEN, MSG_WAITALL);
	close(sock);
	if (ret < 0) {
		if(errno == EWOULDBLOCK)
			return ret;
		log_error("socket_request: could not recv message: %d\n", ret);
		return ret;
	}

	ret = socket_msg_unmarshal(answer, msg);
	if (ret) {
		log_error("socket_request: could not unmarshel message: %d\n", ret);
		return ret;
	}

	log_verbose("socket_request: end %d\n", ret);
	if (answer->type & SOCKET_MSG_TYPE_RESPONSE) {
		return 0;
	}
	return -1;
}

static void server_cb(struct uloop_fd *fd, unsigned int events) {
	struct sockaddr_in6 addr;
	char buf[REQUEST_MAXLEN];
	char control[256];


	struct iovec iv = {
		.iov_base = buf,
		.iov_len = sizeof(buf) - 1
	};

	struct msghdr mh = {
		.msg_name = &addr,
		.msg_namelen = sizeof(addr),
		.msg_iov = &iv,
		.msg_iovlen = 1,
		.msg_control = control,
		.msg_controllen = sizeof(control)
	};

	if (recvmsg(server.fd, &mh, MSG_WAITALL) < 0) {
		log_info("socket: invalid package\n");
		return;
	}

	char addr_str[INET6_ADDRSTRLEN];
	inet_ntop(AF_INET6, &(addr.sin6_addr), addr_str, INET6_ADDRSTRLEN);
	log_info("socket:recv [%s]:%d - %s\n", addr_str, ntohs(addr.sin6_port), buf);


	if (sendto(server.fd, addr_str, sizeof(addr_str) - 1, 0, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		log_error("socket: could not send answer\n");
	}
	log_info("socket: send answer\n");
}

int socket_init(char *ifname) {
	server.cb = server_cb;
	server.fd = usock(USOCK_UDP | USOCK_SERVER | USOCK_IPV6ONLY | USOCK_NUMERIC, SOCKET_ADDR, SOCKET_PORT);
	if (server.fd < 0) {
		return 1;
	}

	client_addr.sin6_family = AF_INET6;
	client_addr.sin6_port = htons(atoi(SOCKET_PORT));
	client_addr.sin6_scope_id = if_nametoindex(ifname);

	// listen multicast
	if (!inet_pton(AF_INET6, SOCKET_MADDR, &client_addr.sin6_addr)) {
		log_error("socket: invalid multicast group\n");
		return -1;
	}

	struct ipv6_mreq mreq = {
		.ipv6mr_multiaddr = client_addr.sin6_addr,
		.ipv6mr_interface = client_addr.sin6_scope_id,
	};

	if (setsockopt(server.fd, IPPROTO_IPV6, IPV6_JOIN_GROUP, &mreq, sizeof(mreq)) == -1) {
		log_error("socket: unable to join multicast group: %d\n", errno);
		return -1;
	}

	int loop = 0;
	if(setsockopt(server.fd, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, &loop, sizeof(loop)) < 0) {
		log_error("socket_request: unable to disable listen/loop own multicast package: %d\n", errno);
		return -1;
	}
	// end listen multicast
	uloop_fd_add(&server, ULOOP_READ);

	return 0;
}

void socket_close() {
	uloop_fd_delete(&server);
	//ustream_free(&(server.fd));
	close(server.fd);
}
